import ast
import re
from typing import Optional
from dataclasses import dataclass, field

from tach import filesystem as fs


@dataclass
class IgnoreDirective:
    lineno: int
    modules: list[str] = field(default_factory=list)


TACH_IGNORE_REGEX = re.compile(r"# *tach-ignore(( [\w.]+)*)$")


def get_ignore_directives(file_content: str) -> dict[int, IgnoreDirective]:
    ignores: dict[int, IgnoreDirective] = {}
    lines = file_content.splitlines()
    for lineno, line in enumerate(lines):
        normal_lineno = lineno + 1
        match = TACH_IGNORE_REGEX.match(line)
        if match:
            ignored_modules = match.group(1)
            if ignored_modules:
                ignores[normal_lineno] = IgnoreDirective(
                    lineno=lineno + 1, modules=ignored_modules.split()
                )
            else:
                ignores[normal_lineno] = IgnoreDirective(lineno=lineno + 1)
    return ignores


class ImportVisitor(ast.NodeVisitor):
    def __init__(
        self,
        project_root: str,
        current_mod_path: str,
        is_package: bool = False,
        ignore_directives: Optional[dict[int, IgnoreDirective]] = None,
    ):
        self.project_root = project_root
        self.current_mod_path = current_mod_path
        self.is_package = is_package
        self.ignored_imports = ignore_directives or {}
        self.imports: list[str] = []

    def _get_ignored_modules(self, lineno: int) -> Optional[list[str]]:
        # Check for ignore directive at the previous line or on the current line
        directive = self.ignored_imports.get(lineno - 1) or self.ignored_imports.get(
            lineno
        )
        return directive.modules if directive else None

    def visit_ImportFrom(self, node: ast.ImportFrom):
        # For relative imports (level > 0), adjust the base module path
        if node.module is not None and node.level > 0:
            num_paths_to_strip = node.level - 1 if self.is_package else node.level
            base_path_parts = self.current_mod_path.split(".")
            if num_paths_to_strip:
                base_path_parts = base_path_parts[:-num_paths_to_strip]
            base_mod_path = ".".join(
                [*base_path_parts, node.module if node.module else ""]
            )
        else:
            base_mod_path = node.module or ""

        ignored_modules = self._get_ignored_modules(node.lineno)

        if ignored_modules is not None and len(ignored_modules) == 0:
            # Empty ignore list signifies blanket ignore of following import
            return

        for name_node in node.names:
            local_mod_path = f"{'.' * node.level}{node.module or ''}.{name_node.asname or name_node.name}"
            if ignored_modules is not None and (local_mod_path in ignored_modules):
                # This import is ignored by a tach-ignore directive
                continue

            global_mod_path = (
                f"{base_mod_path}.{name_node.name}" if node.module else name_node.name
            )
            if fs.is_project_import(self.project_root, global_mod_path):
                self.imports.append(global_mod_path)

    def visit_Import(self, node: ast.Import):
        ignored_modules = self._get_ignored_modules(node.lineno)
        if ignored_modules is not None and len(ignored_modules) == 0:
            # Empty ignore list signifies blanket ignore of following import
            return

        ignored_modules = ignored_modules or []
        for alias in node.names:
            if alias.name in ignored_modules or not fs.is_project_import(
                self.project_root, alias.name
            ):
                continue
            self.imports.append(alias.name)


def get_project_imports(project_root: str, file_path: str) -> list[str]:
    file_content = fs.read_file(file_path)
    parsed_ast = fs.parse_ast(file_path)
    ignore_directives = get_ignore_directives(file_content)
    mod_path = fs.file_to_module_path(file_path)
    import_visitor = ImportVisitor(
        project_root=project_root,
        is_package=file_path.endswith("__init__.py"),
        current_mod_path=mod_path,
        ignore_directives=ignore_directives,
    )
    import_visitor.visit(parsed_ast)
    return import_visitor.imports



from tach.parsing.config import parse_package_config, parse_project_config
from tach.parsing.imports import get_project_imports
from tach.parsing.interface import parse_interface_members
from tach.parsing.packages import build_package_trie


__all__ = [
    "parse_package_config",
    "parse_project_config",
    "get_project_imports",
    "parse_interface_members",
    "build_package_trie",
]



import pytest

from tach.core import PackageTrie, PackageNode, PackageConfig


@pytest.fixture
def test_config() -> PackageConfig:
    return PackageConfig(tags=["test"], strict=False)


@pytest.fixture
def package_trie() -> PackageTrie:
    return PackageTrie(
        root=PackageNode(
            is_end_of_path=False,
            full_path="",
            config=None,
            children={
                "domain_one": PackageNode(
                    is_end_of_path=True,
                    full_path="domain_one",
                    config=PackageConfig(tags=["test"], strict=False),
                    children={
                        "subdomain": PackageNode(
                            is_end_of_path=True,
                            full_path="domain_one.subdomain",
                            config=PackageConfig(tags=["test"], strict=False),
                            children={},
                        )
                    },
                ),
                "domain_two": PackageNode(
                    is_end_of_path=True,
                    full_path="domain_two",
                    config=PackageConfig(tags=["test"], strict=False),
                    children={
                        "subdomain": PackageNode(
                            is_end_of_path=True,
                            full_path="domain_two.subdomain",
                            config=PackageConfig(tags=["test"], strict=False),
                            children={},
                        )
                    },
                ),
                "domain_three": PackageNode(
                    is_end_of_path=True,
                    full_path="domain_three",
                    config=PackageConfig(tags=["test"], strict=False),
                    children={},
                ),
            },
        )
    )


def test_iterate_over_empty_trie():
    assert list(PackageTrie()) == []


def test_iterate_over_populated_trie(package_trie):
    assert set((node.full_path for node in package_trie)) == {
        "domain_one",
        "domain_one.subdomain",
        "domain_two",
        "domain_two.subdomain",
        "domain_three",
    }


def test_get_nonexistent_path(package_trie):
    assert package_trie.get("fakepath") is None


def test_get_nonexistent_empty_path():
    trie = PackageTrie()
    assert trie.get("") is None


def test_get_actual_path(package_trie):
    assert package_trie.get("domain_one") is not None


def test_insert_empty_path(test_config):
    trie = PackageTrie()
    trie.insert(test_config, "", [])
    assert set((node.full_path for node in trie)) == {""}


def test_insert_single_level_path(test_config):
    trie = PackageTrie()
    trie.insert(test_config, "domain", [])
    assert set((node.full_path for node in trie)) == {"domain"}


def test_insert_multi_level_path(test_config):
    trie = PackageTrie()
    trie.insert(test_config, "domain.subdomain", [])
    assert set((node.full_path for node in trie)) == {"domain.subdomain"}


def test_find_nearest_at_root(package_trie):
    package = package_trie.find_nearest("other_domain")
    assert package is None


def test_find_nearest_in_single_domain(package_trie):
    package = package_trie.find_nearest("domain_one.thing")
    assert package.full_path == "domain_one"


def test_find_nearest_in_nested_domain(package_trie):
    package = package_trie.find_nearest("domain_two.subdomain.thing")
    assert package.full_path == "domain_two.subdomain"



import pytest
import tempfile
import shutil
import os
from tach import errors, filesystem as fs
from tach.init import init_project


def init_project_from_root(root) -> None:
    # Save the current working directory
    saved_directory = os.getcwd()
    try:
        # Navigate to the root directory and call init_project
        fs.chdir(root)
        init_project(root)
    finally:
        # Change back to the original directory
        fs.chdir(saved_directory)


@pytest.fixture(scope="module")
def test_root():
    # Create a temporary directory to use as the root for testing
    test_root = tempfile.mkdtemp()
    yield test_root
    # Remove the temporary directory after testing
    shutil.rmtree(test_root)


def test_init_project_with_valid_root(test_root):
    # Create some mock files and directories for testing
    test_dirs = [
        "package1",
        "package2",
        "package3",
        "package4/subpackage",
        "package5/subpackage",
        "package6/subpackage",
    ]
    for d in test_dirs:
        os.makedirs(os.path.join(test_root, d))
        with open(os.path.join(test_root, d, "__init__.py"), "w") as f:
            f.write("# Mock __init__.py file")

    # Create some mock Python files with imports and member names
    file_contents = {
        "package1/__init__.py": "from package6.subpackage.module6 import x",
        "package2/__init__.py": "",
        "package1/module1.py": "class Package1Class:\n    pass\n",
        "package2/module2.py": "def package_2_func():\n    pass\n",
        "package3/__init__.py": "from package1.module1 import Package1Class\nfrom package2.module2 import package_2_func\n",
        "package3/module3.py": "",
        "package4/subpackage/__init__.py": "",
        "package5/subpackage/__init__.py": "import package3.module3",
        "package6/subpackage/__init__.py": "",
        "package6/subpackage/module6.py": "x = 3\n",
    }

    for file_path, content in file_contents.items():
        with open(os.path.join(test_root, file_path), "w") as f:
            f.write(content)

    # Call init_project with the test root
    init_project_from_root(test_root)

    # TODO: test new behavior


def test_init_project_with_invalid_root():
    with pytest.raises(errors.TachSetupError):
        init_project("nonexistent_directory")



from tach.core.config import PackageConfig, ScopeDependencyRules, ProjectConfig
from tach.core.package import PackageNode, PackageTrie

__all__ = [
    "PackageConfig",
    "ScopeDependencyRules",
    "ProjectConfig",
    "PackageNode",
    "PackageTrie",
]



import ast
import os
from typing import Any

from tach import filesystem as fs
from tach.parsing.ast_visitor import EarlyExitNodeVisitor


class InterfaceVisitor(EarlyExitNodeVisitor):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.members: list[str] = []

    def visit_Assign(self, node: ast.Assign):
        # Check if the node is an assignment
        if any(
            isinstance(target, ast.Name) and target.id == "__all__"
            for target in node.targets
        ):
            # Ensure the value is a list and process each element
            if isinstance(node.value, ast.List):
                for element in node.value.elts:
                    if isinstance(element, ast.Str):  # pyright: ignore
                        # Supports Python 3.7 and below
                        self.members.append(element.s)
                    elif isinstance(element, ast.Constant):
                        # Supports Python 3.8 and above
                        if isinstance(element.value, str):
                            self.members.append(element.value)
                # Early exit
                self.set_exit(True)
        # Continue to the next node
        self.generic_visit(node)


def parse_interface_members(path: str) -> list[str]:
    """
    Parse the members of __all__ in the __init__.py of a package.
    'path' should be a path to the package directory.
    """
    init_py_path = os.path.join(path, "__init__.py")
    if not os.path.exists(init_py_path):
        raise ValueError(
            f"Could not parse interface from path, no __init__.py found: {path}"
        )
    parsed_ast = fs.parse_ast(init_py_path)
    interface_visitor = InterfaceVisitor()
    interface_visitor.visit(parsed_ast)
    return interface_visitor.members



import threading
import sys
import itertools
import queue
import time

SPINNER_CHARS = ".oOo"
spinner_started = False
SPINNER_CHAR_FPS = 3
SPINNER_CHAR_TIME_DELAY = 1 / SPINNER_CHAR_FPS
SPINNER_DELAY = 0.35


end_signal_queue: queue.Queue[bool] = queue.Queue()
confirm_end_signal_queue: queue.Queue[bool] = queue.Queue()


def spinner(label: str = ""):
    time.sleep(SPINNER_DELAY)
    written = False
    for spinner_char in itertools.cycle(SPINNER_CHARS):
        line = f"{spinner_char} {label}" if label else spinner_char
        try:
            end_signal_queue.get(timeout=SPINNER_CHAR_TIME_DELAY)
            if written:
                sys.stdout.write("\b" * len(line))
                sys.stdout.flush()
            confirm_end_signal_queue.put_nowait(True)
            return
        except queue.Empty:
            pass
        sys.stdout.write(line)
        sys.stdout.flush()
        written = True
        sys.stdout.write("\b" * len(line))


def stop_spinner():
    global spinner_started
    if spinner_started:
        end_signal_queue.put_nowait(True)
        try:
            confirm_end_signal_queue.get(timeout=SPINNER_CHAR_TIME_DELAY + 0.2)
        except queue.Empty:
            pass
        spinner_started = False


def start_spinner(label: str = ""):
    global spinner_started
    if not spinner_started:
        threading.Thread(target=spinner, kwargs={"label": label}, daemon=True).start()
        spinner_started = True



import os
from dataclasses import field, dataclass
from typing import Optional

import yaml

from tach import errors
from tach import filesystem as fs
from tach.check import check
from tach.constants import PACKAGE_FILE_NAME, CONFIG_FILE_NAME
from tach.core import ProjectConfig, ScopeDependencyRules

__package_yml_template = """tags: ['{dir_name}']\n"""


@dataclass
class PackageInitResult:
    package_paths: list[str] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)


def init_packages(
    root: str, depth: int, exclude_paths: Optional[list[str]] = None
) -> PackageInitResult:
    package_paths: list[str] = []
    warnings: list[str] = []
    for dir_path in fs.walk_pypackages(root, depth=depth, exclude_paths=exclude_paths):
        package_yml_path = os.path.join(dir_path, f"{PACKAGE_FILE_NAME}.yml")
        package_paths.append(dir_path)
        if os.path.exists(package_yml_path):
            warnings.append(f"Package file '{package_yml_path}' already exists.")
            continue
        package_yml_content = __package_yml_template.format(
            dir_name=dir_path.replace(os.path.sep, ".")
        )
        fs.write_file(package_yml_path, package_yml_content)

    return PackageInitResult(package_paths=package_paths, warnings=warnings)


@dataclass
class InitRootResult:
    warnings: list[str] = field(default_factory=list)


def init_root(root: str, exclude_paths: Optional[list[str]] = None) -> InitRootResult:
    project_config_path = fs.get_project_config_path(root)
    if project_config_path:
        return InitRootResult(
            warnings=[f"Project already contains {CONFIG_FILE_NAME}.yml"]
        )

    project_config = ProjectConfig()
    check_errors = check(
        root, project_config=project_config, exclude_paths=exclude_paths
    )
    for error in check_errors:
        if error.is_tag_error:
            existing_dependencies = set(
                project_config.constraints.get(
                    error.source_tag, ScopeDependencyRules(depends_on=[])
                ).depends_on
            )
            project_config.constraints[error.source_tag] = ScopeDependencyRules(
                depends_on=list(existing_dependencies | set(error.invalid_tags))
            )

    tach_yml_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
    tach_yml_content = yaml.dump(project_config.model_dump())
    fs.write_file(tach_yml_path, tach_yml_content)

    check_errors = check(
        root, project_config=project_config, exclude_paths=exclude_paths
    )
    if check_errors:
        return InitRootResult(
            warnings=[
                "Could not auto-detect all dependencies, use 'tach check' to finish initialization manually."
            ]
        )

    return InitRootResult(warnings=[])


def init_project(
    root: str, depth: Optional[int] = None, exclude_paths: Optional[list[str]] = None
) -> list[str]:
    if not os.path.isdir(root):
        raise errors.TachSetupError(f"The path {root} is not a directory.")

    if exclude_paths is None:
        exclude_paths = ["tests/", "docs/"]

    warnings: list[str] = []

    if depth is None:
        package_init_result = init_packages(root, depth=1, exclude_paths=exclude_paths)
        warnings.extend(package_init_result.warnings)
        if len(package_init_result.package_paths) == 1:
            result = init_packages(
                package_init_result.package_paths[0],
                depth=1,
                exclude_paths=exclude_paths,
            )
            warnings.extend(result.warnings)
    else:
        package_init_result = init_packages(
            root, depth=depth, exclude_paths=exclude_paths
        )
        warnings.extend(package_init_result.warnings)

    init_root_result = init_root(root, exclude_paths=exclude_paths)
    warnings.extend(init_root_result.warnings)

    return warnings



import os
from typing import Optional


from tach.constants import PACKAGE_FILE_NAME


def validate_package_config(root: str = ".") -> Optional[str]:
    file_path = os.path.join(root, f"{PACKAGE_FILE_NAME}.yml")
    if os.path.exists(file_path):
        return file_path
    file_path = os.path.join(root, f"{PACKAGE_FILE_NAME}.yaml")
    if os.path.exists(file_path):
        return file_path
    return



import pytest
from unittest.mock import patch, MagicMock

from tach.constants import CONFIG_FILE_NAME
from tach.errors import TachError
from tach.add import validate_path


def test_path_does_not_exist():
    with pytest.raises(TachError) as excinfo:
        validate_path("/non/existent/path")
    assert "/non/existent/path does not exist." in str(excinfo.value)


def test_path_is_directory_with_package_file():
    with patch(
        "tach.filesystem.package.os.path.exists",
        MagicMock(side_effect=lambda x: x.endswith(".yml") or x.endswith("dir")),
    ), patch("tach.filesystem.package.os.path.isdir", return_value=True):
        with pytest.raises(TachError) as excinfo:
            validate_path("/some/dir")
    assert "/some/dir already contains a package.yml" in str(excinfo.value)


def test_directory_without_init_py():
    def mock_exists(path):
        return path.endswith("dir") and not path.endswith("__init__.py")

    with patch("tach.filesystem.package.os.path.exists", mock_exists), patch(
        "tach.filesystem.package.os.path.isdir", return_value=True
    ):
        with pytest.raises(TachError) as excinfo:
            validate_path("/some/dir")
    assert "/some/dir is not a valid Python package (no __init__.py found)." in str(
        excinfo.value
    )


def test_valid_directory():
    def mock_exists(path):
        return (
            not path.endswith("yml")
            and not path.endswith("yaml")
            or CONFIG_FILE_NAME in path
        )  # Everything exists for this test

    def mock_validate_project_config_path(path):
        return  # Assume validation is successful

    with patch("tach.filesystem.package.os.path.exists", mock_exists), patch(
        "tach.filesystem.validate_project_config_path",
        mock_validate_project_config_path,
    ), patch("tach.filesystem.package.os.path.isdir", return_value=True):
        # No exception should be raised
        validate_path("/some/dir")


def test_non_python_file():
    with patch("tach.filesystem.package.os.path.exists", return_value=True):
        with pytest.raises(TachError) as excinfo:
            validate_path("/some/file.txt")
    assert "/some/file.txt is not a Python file." in str(excinfo.value)


def test_python_file_with_matching_directory():
    def mock_exists(path):
        if path.endswith(".py"):
            return True
        return path == "/some/file"

    with patch("tach.filesystem.package.os.path.exists", mock_exists):
        with pytest.raises(TachError) as excinfo:
            validate_path("/some/file.py")
    assert "{path} already has a directory of the same name." in str(excinfo.value)


def test_valid_python_file():
    def mock_exists(path):
        return path.endswith(".py")

    with patch(
        "tach.filesystem.package.os.path.exists", side_effect=mock_exists
    ), patch("tach.filesystem.validate_project_config_path", side_effect=SystemError):
        with pytest.raises(TachError) as excinfo:
            validate_path("/some/file.py")
    assert f"{CONFIG_FILE_NAME} does not exist in any parent directories" in str(
        excinfo.value
    )



import argparse
import sys
from typing import Optional

from tach.add import add_packages
from tach.check import check, ErrorInfo
from tach import filesystem as fs
from tach.constants import CONFIG_FILE_NAME
from tach.init import init_project
from tach.loading import stop_spinner, start_spinner
from tach.parsing import parse_project_config
from tach.colors import BCOLORS


def print_errors(error_list: list[ErrorInfo]) -> None:
    sorted_results = sorted(error_list, key=lambda e: e.location)
    for error in sorted_results:
        print(
            f"❌ {BCOLORS.FAIL}{error.location}{BCOLORS.WARNING}: {error.message}",
            file=sys.stderr,
        )


def add_base_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-e",
        "--exclude",
        required=False,
        type=str,
        metavar="file_or_path,...",
        help="Comma separated path list to exclude. tests/, ci/, etc.",
    )


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="tach",
        add_help=True,
        epilog="Make sure tach is run from the root of your Python project,"
        " and `tach.yml` is present",
    )
    subparsers = parser.add_subparsers(title="commands", dest="command")
    init_parser = subparsers.add_parser(
        "init",
        prog="tach init",
        help="Initialize boundaries between top-level packages and write dependencies to "
        "`tach.yml`",
        description="Initialize boundaries between top-level packages and write dependencies to "
        "`tach.yml`",
    )
    init_parser.add_argument(
        "-d",
        "--depth",
        type=int,
        nargs="?",
        default=None,
        help="The number of child directories to search for packages to initialize",
    )
    add_base_arguments(init_parser)
    check_parser = subparsers.add_parser(
        "check",
        prog="tach check",
        help="Check existing boundaries against your dependencies and package interfaces",
        description="Check existing boundaries against your dependencies and package interfaces",
    )
    add_base_arguments(check_parser)
    add_parser = subparsers.add_parser(
        "add",
        prog="tach add",
        help="Create a new module boundary around an existing file or folder",
        description="Initialize boundaries between top-level modules and write dependencies to "
        "`tach.yml`",
    )
    add_parser.add_argument(
        "path",
        type=str,
        metavar="file_or_path,...",
        help="The path(s) of the file or directory to create a module boundary around. "
        "Use a comma-separated list for multiple.",
    )
    add_parser.add_argument(
        "-t",
        "--tags",
        required=False,
        type=str,
        metavar="tag,...",
        help="The tag for the module to be initialized with."
        "Use a comma-separated list for multiple.",
    )
    return parser


def parse_arguments(
    args: list[str],
) -> tuple[argparse.Namespace, argparse.ArgumentParser]:
    parser = build_parser()
    parsed_args = parser.parse_args(args)

    if args[0] not in ["init", "add"]:
        fs.validate_project_config_path()

    return parsed_args, parser


def tach_check(
    exclude_paths: Optional[list[str]] = None,
):
    try:
        project_config = parse_project_config()
        if exclude_paths is not None and project_config.exclude is not None:
            exclude_paths.extend(project_config.exclude)
        else:
            exclude_paths = project_config.exclude
        result: list[ErrorInfo] = check(
            ".",
            project_config,
            exclude_paths=exclude_paths,
            exclude_hidden_paths=project_config.exclude_hidden_paths,
        )
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)

    stop_spinner()
    if result:
        print_errors(result)
        sys.exit(1)
    print(f"✅ {BCOLORS.OKGREEN}All package dependencies validated!")
    sys.exit(0)


def tach_init(depth: Optional[int] = None, exclude_paths: Optional[list[str]] = None):
    try:
        warnings = init_project(root=".", depth=depth, exclude_paths=exclude_paths)
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)

    stop_spinner()
    if warnings:
        print("\n".join(warnings))
    print(f"✅ {BCOLORS.OKGREEN}Initialized {CONFIG_FILE_NAME}.yml.")
    sys.exit(0)


def tach_add(paths: set[str], tags: Optional[set[str]] = None) -> None:
    try:
        warnings = add_packages(paths, tags)
    except Exception as e:
        stop_spinner()
        print(str(e))
        sys.exit(1)

    stop_spinner()
    if warnings:
        print("\n".join(warnings))
    if len(paths) > 1:
        print(f"✅ {BCOLORS.OKGREEN}Packages added.")
    else:
        print(f"✅ {BCOLORS.OKGREEN}Package added.")
    sys.exit(0)


def main() -> None:
    args, parser = parse_arguments(sys.argv[1:])
    if args.command == "add":
        paths = set(args.path.split(","))
        tags = set(args.tags.split(",")) if args.tags else None
        tach_add(paths=paths, tags=tags)
        return
    exclude_paths = args.exclude.split(",") if args.exclude else None
    if args.command == "init":
        start_spinner("Initializing...")
        tach_init(depth=args.depth, exclude_paths=exclude_paths)
    elif args.command == "check":
        start_spinner("Scanning...")
        tach_check(exclude_paths=exclude_paths)
    else:
        print("Unrecognized command")
        parser.print_help()
        exit(1)


if __name__ == "__main__":
    main()



from typing import List, Dict, Optional

from pydantic import BaseModel, Field


class Config(BaseModel):
    model_config = {"extra": "forbid"}


class PackageConfig(Config):
    """
    Configuration for a single package within a project.
    """

    tags: List[str]
    strict: bool = False


class ScopeDependencyRules(Config):
    """
    Dependency rules for a particular scope.
    """

    depends_on: List[str]


class ProjectConfig(Config):
    """
    Configuration applied globally to a project.
    """

    constraints: Dict[str, ScopeDependencyRules] = Field(default_factory=dict)
    exclude: Optional[List[str]] = Field(default_factory=lambda: ["tests", "docs"])
    exclude_hidden_paths: Optional[bool] = True



from typing import Optional

from tach import filesystem as fs
from tach.core import PackageTrie
from tach.parsing import parse_package_config, parse_interface_members


def build_package_trie(
    root: str,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
) -> PackageTrie:
    package_trie = PackageTrie()

    for dir_path in fs.walk_configured_packages(
        root,
        exclude_paths=exclude_paths,
        exclude_hidden_paths=exclude_hidden_paths,
    ):
        package_config = parse_package_config(dir_path)
        if package_config is None:
            raise ValueError(f"Could not parse package config for {dir_path}")
        package_trie.insert(
            config=package_config,
            path=fs.file_to_module_path(dir_path),
            interface_members=parse_interface_members(dir_path),
        )

    return package_trie



from typing import Optional

import yaml

from tach.core import ProjectConfig, PackageConfig
from tach import filesystem as fs


def parse_project_config(root: str = ".") -> ProjectConfig:
    file_path = fs.validate_project_config_path(root)
    with open(file_path, "r") as f:
        result = yaml.safe_load(f)
        if not result or not isinstance(result, dict):
            raise ValueError(f"Empty or invalid project config file: {file_path}")
    # We want to error on type issues here for now
    project_config = ProjectConfig(**result)  # type: ignore
    return project_config


def parse_package_config(root: str = ".") -> Optional[PackageConfig]:
    file_path = fs.validate_package_config(root)
    if file_path:
        with open(file_path, "r") as f:
            result = yaml.safe_load(f)
            if not result or not isinstance(result, dict):
                raise ValueError(f"Empty or invalid package config file: {file_path}")
        # We want to error on type issues here for now
        return PackageConfig(**result)  # type: ignore



from unittest.mock import Mock
import pytest

from tach import cli
from tach.check import ErrorInfo
from tach.constants import CONFIG_FILE_NAME
from tach.core import ProjectConfig


@pytest.fixture
def mock_check(mocker) -> Mock:
    mock = Mock(return_value=[])  # default to a return with no errors
    mocker.patch("tach.cli.check", mock)
    return mock


@pytest.fixture
def mock_isdir(mocker) -> None:
    def mock_isdir(path: str) -> bool:
        if path == "valid_dir":
            return True
        else:
            return False

    mocker.patch("tach.filesystem.project.os.path.isdir", mock_isdir)


@pytest.fixture
def mock_path_exists(mocker) -> None:
    def mock_path_exists(path: str) -> bool:
        if CONFIG_FILE_NAME in path:
            return True
        else:
            return False

    mocker.patch("tach.filesystem.project.os.path.exists", mock_path_exists)


@pytest.fixture
def mock_project_config(mocker) -> None:
    def mock_project_config() -> ProjectConfig:
        return ProjectConfig()

    mocker.patch("tach.cli.parse_project_config", mock_project_config)


def test_execute_with_tach_yml(
    capfd, mock_path_exists, mock_check, mock_project_config
):
    # Test with a valid path as mocked
    args, _ = cli.parse_arguments(["check"])
    assert args.command == "check"
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check()
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All package dependencies validated!" in captured.out


def test_execute_with_error(capfd, mock_path_exists, mock_check, mock_project_config):
    # Mock an error returned from check
    location = "valid_dir/file.py"
    message = "Import valid_dir in valid_dir/file.py is blocked by boundary"
    mock_check.return_value = [
        ErrorInfo(
            exception_message="Import valid_dir in valid_dir/file.py is blocked by boundary",
        )
    ]
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check()
    captured = capfd.readouterr()
    assert sys_exit.value.code == 1
    assert location in captured.err
    assert message in captured.err


def test_execute_with_no_tach_yml(capfd):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with no tach.yml mocked
        cli.parse_arguments(["check"])
    captured = capfd.readouterr()
    assert sys_exit.value.code == 1
    assert "tach.(yml|yaml) not found" in captured.err


def test_invalid_command(capfd):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with an invalid command
        cli.parse_arguments(["help"])
    captured = capfd.readouterr()
    assert sys_exit.value.code == 2
    assert "invalid choice: 'help" in captured.err


def test_execute_with_valid_exclude(
    capfd, mock_isdir, mock_path_exists, mock_check, mock_project_config
):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with a valid path as mocked
        args, _ = cli.parse_arguments(["check", "--exclude", "valid_dir"])
        exclude_paths = args.exclude.split(",")
        cli.tach_check(exclude_paths=exclude_paths)
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All package dependencies validated!" in captured.out



import os
import re
from dataclasses import dataclass, field
from typing import Optional

from tach import filesystem as fs
from tach.core import PackageTrie, PackageNode, ProjectConfig
from tach.parsing import build_package_trie, get_project_imports


@dataclass
class ErrorInfo:
    location: str = ""
    import_mod_path: str = ""
    source_tag: str = ""
    invalid_tags: list[str] = field(default_factory=list)
    allowed_tags: list[str] = field(default_factory=list)
    exception_message: str = ""

    @property
    def is_tag_error(self) -> bool:
        return all(
            (self.location, self.import_mod_path, self.source_tag, self.invalid_tags)
        )

    @property
    def message(self) -> str:
        if self.exception_message:
            return self.exception_message
        if not self.is_tag_error:
            return f"Unexpected error: ({[self.location, self.import_mod_path, self.source_tag, self.allowed_tags]})"
        if not self.allowed_tags:
            return (
                f"Import '{self.import_mod_path}' with tags '{self.invalid_tags}' "
                f"in {self.location} is blocked. "
                f"Tag '{self.source_tag}' has no allowed dependency tags."
            )
        return (
            f"Import '{self.import_mod_path}' with tags '{self.invalid_tags}' "
            f"in {self.location} is blocked. "
            f"Tag '{self.source_tag}' can only depend on tags '{self.allowed_tags}'."
        )


@dataclass
class CheckResult:
    ok: bool
    error_info: Optional[ErrorInfo] = None

    @classmethod
    def success(cls) -> "CheckResult":
        return cls(ok=True)

    @classmethod
    def fail(cls, error_info: ErrorInfo) -> "CheckResult":
        return cls(ok=False, error_info=error_info)


def is_top_level_package_import(mod_path: str, package: PackageNode) -> bool:
    return mod_path == package.full_path


def import_matches_interface_members(mod_path: str, package: PackageNode) -> bool:
    mod_path_basename = mod_path.rsplit(".", 1)[-1]
    return mod_path_basename in package.interface_members


def check_import(
    project_config: ProjectConfig,
    package_trie: PackageTrie,
    import_mod_path: str,
    file_mod_path: str,
    file_nearest_package: Optional[PackageNode] = None,
) -> CheckResult:
    import_nearest_package = package_trie.find_nearest(import_mod_path)
    if import_nearest_package is None:
        # This shouldn't happen since we intend to filter out any external imports,
        # but we should allow external imports if they have made it here.
        return CheckResult.success()

    # Lookup file_mod_path if package not given
    if file_nearest_package is None:
        file_nearest_package = package_trie.find_nearest(file_mod_path)
    # If package not found, we should fail since the implication is that
    # an external package is importing directly from our project
    if file_nearest_package is None:
        return CheckResult.fail(
            error_info=ErrorInfo(
                exception_message=f"Package containing '{file_mod_path}' not found in project."
            )
        )

    # Imports within the same package are always allowed
    if import_nearest_package == file_nearest_package:
        return CheckResult.success()

    import_package_config = import_nearest_package.config
    if import_package_config and import_package_config.strict:
        if not is_top_level_package_import(
            import_mod_path, import_nearest_package
        ) and not import_matches_interface_members(
            import_mod_path, import_nearest_package
        ):
            # In strict mode, import must be of the package itself or one of the
            # interface members (defined in __all__)
            return CheckResult.fail(
                error_info=ErrorInfo(
                    location=file_mod_path,
                    exception_message=(
                        f"Package '{import_nearest_package.full_path}' is in strict mode. "
                        "Only imports from the root of this package are allowed. "
                        f"The import '{import_mod_path}' (in '{file_mod_path}') "
                        f"is not included in __all__."
                    ),
                )
            )

    # The import must be explicitly allowed based on the tags and top-level config
    if not file_nearest_package.config or not import_nearest_package.config:
        return CheckResult.fail(
            error_info=ErrorInfo(
                exception_message="Could not find config for packages."
            )
        )
    file_tags = file_nearest_package.config.tags
    import_tags = import_nearest_package.config.tags

    for file_tag in file_tags:
        dependency_tags = (
            project_config.constraints[file_tag].depends_on
            if file_tag in project_config.constraints
            else []
        )
        if any(
            any(
                re.match(dependency_tag, import_tag)
                for dependency_tag in dependency_tags
            )
            for import_tag in import_tags
        ):
            # The import has at least one tag which matches at least one expected dependency
            continue
        # This means the import has no tags which the file can depend on
        return CheckResult.fail(
            error_info=ErrorInfo(
                location=file_mod_path,
                import_mod_path=import_mod_path,
                source_tag=file_tag,
                invalid_tags=import_tags,
                allowed_tags=dependency_tags,
            )
        )

    return CheckResult.success()


def check(
    root: str,
    project_config: ProjectConfig,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
) -> list[ErrorInfo]:
    if not os.path.isdir(root):
        return [
            ErrorInfo(exception_message=f"The path {root} is not a valid directory.")
        ]

    # This 'canonicalizes' the path arguments, resolving directory traversal
    root = fs.canonical(root)
    exclude_paths = list(map(fs.canonical, exclude_paths)) if exclude_paths else None

    package_trie = build_package_trie(
        root, exclude_paths=exclude_paths, exclude_hidden_paths=exclude_hidden_paths
    )

    errors: list[ErrorInfo] = []
    for file_path in fs.walk_pyfiles(
        root, exclude_paths=exclude_paths, exclude_hidden_paths=exclude_hidden_paths
    ):
        mod_path = fs.file_to_module_path(file_path)
        nearest_package = package_trie.find_nearest(mod_path)
        if nearest_package is None:
            continue
        import_mod_paths = get_project_imports(root, file_path)
        # This should only give us imports from within our project
        # (excluding stdlib, builtins, and 3rd party packages)
        for import_mod_path in import_mod_paths:
            check_result = check_import(
                project_config=project_config,
                package_trie=package_trie,
                import_mod_path=import_mod_path,
                file_nearest_package=nearest_package,
                file_mod_path=mod_path,
            )
            if check_result.ok or check_result.error_info is None:
                continue

            errors.append(check_result.error_info)

    return errors



import os
import sys
from pathlib import Path
from typing import Optional

from tach.colors import BCOLORS
from tach.constants import CONFIG_FILE_NAME


def print_no_config_yml() -> None:
    print(
        f"{BCOLORS.FAIL} {CONFIG_FILE_NAME}.(yml|yaml) not found in {os.getcwd()}",
        file=sys.stderr,
    )


def get_project_config_path(root: str = ".") -> str:
    file_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
    if os.path.exists(file_path):
        return file_path
    file_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yaml")
    if os.path.exists(file_path):
        return file_path
    return ""


def find_project_config_root(path: str) -> Optional[str]:
    path = os.path.abspath(path)
    if os.path.isdir(path):
        if get_project_config_path(path):
            return path
    path_obj = Path(path)
    # Iterate upwards, looking for project config
    for parent in path_obj.parents:
        if get_project_config_path(str(parent)):
            return str(parent)


def validate_project_config_path(root: str = ".") -> str:
    project_config_path = get_project_config_path(root)
    if not project_config_path:
        print_no_config_yml()
        sys.exit(1)
    else:
        return project_config_path



import os
import ast
import re
import sys
import threading
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Optional, Generator
from tach import errors


@dataclass
class FileInfo:
    path: str
    content: Optional[str] = None
    canonical_path: Optional[str] = None
    ast: Optional["ast.AST"] = None


# Thread-local file cache to avoid going to disk as much as possible
thread_local = threading.local()
# Cannot type-hint non-self attributes (https://github.com/python/mypy/issues/2388)
# cwd: str
thread_local.cwd = os.getcwd()
# file_caches_by_cwd: defaultdict[str, dict[str, FileInfo]]
thread_local.file_caches_by_cwd = defaultdict(dict)


def get_cwd() -> str:
    # Use a cached cwd to avoid system calls
    if not hasattr(thread_local, "cwd"):
        thread_local.cwd = os.getcwd()
    return thread_local.cwd


def chdir(path: str):
    # When using chdir, update the cached version
    os.chdir(path)
    thread_local.cwd = os.getcwd()


def _get_file_cache() -> dict[str, FileInfo]:
    if not hasattr(thread_local, "file_caches_by_cwd"):
        thread_local.file_caches_by_cwd = defaultdict(dict)
    file_caches_by_cwd: defaultdict[str, dict[str, FileInfo]] = (
        thread_local.file_caches_by_cwd
    )  # type: ignore
    return file_caches_by_cwd[get_cwd()]


def _file_cache_key(path: str) -> str:
    return f"{get_cwd()}:::{path}"


def _cached_file(path: str) -> Optional[FileInfo]:
    return _get_file_cache().get(_file_cache_key(path))


def _set_cached_file(path: str, file_info: FileInfo):
    _get_file_cache()[_file_cache_key(path)] = file_info


def canonical(path: str) -> str:
    cached_file = _cached_file(path)
    if cached_file and cached_file.canonical_path:
        return cached_file.canonical_path

    result = os.path.relpath(os.path.realpath(path), start=get_cwd())

    if cached_file:
        cached_file.canonical_path = result
    else:
        _set_cached_file(path, FileInfo(path=path, canonical_path=result))

    return result


def read_file(path: str) -> str:
    cached_file = _cached_file(path)
    if cached_file and cached_file.content:
        return cached_file.content

    with open(path, "r") as f:
        content = f.read()

    if cached_file:
        cached_file.content = content
    else:
        _set_cached_file(path, FileInfo(path=path, content=content))

    return content


def write_file(path: str, content: str):
    with open(path, "w") as f:
        f.write(content)

    cached_file = _cached_file(path)
    if cached_file:
        cached_file.content = content
        cached_file.ast = None
    else:
        _set_cached_file(path, FileInfo(path=path, content=content))


def parse_ast(path: str) -> ast.AST:
    cached_file = _cached_file(path)
    if cached_file and cached_file.ast:
        return cached_file.ast

    if cached_file and cached_file.content:
        content = cached_file.content
        try:
            ast_result = ast.parse(cached_file.content)
        except SyntaxError as e:
            raise errors.TachParseError(f"Syntax error in {path}: {e}")
    else:
        with open(path, "r") as f:
            content = f.read()
        try:
            ast_result = ast.parse(content)
        except SyntaxError as e:
            raise errors.TachParseError(f"Syntax error in {path}: {e}")

    if cached_file:
        cached_file.content = content
        cached_file.ast = ast_result
    else:
        _set_cached_file(path, FileInfo(path=path, content=content, ast=ast_result))

    return ast_result


def walk_pyfiles(
    root: str,
    depth: Optional[int] = None,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
) -> Generator[str, None, None]:
    canonical_root = canonical(root)
    base_depth = 0 if canonical_root == "." else canonical_root.count(os.path.sep) + 1
    for dirpath, _, filenames in os.walk(canonical_root):
        dirpath = canonical(dirpath)
        dirpath_for_matching = f"{dirpath}/"

        if dirpath == canonical_root:
            continue

        if exclude_hidden_paths and os.path.basename(dirpath).startswith("."):
            continue

        if exclude_paths is not None and any(
            dirpath_for_matching.startswith(exclude_path)
            or re.match(exclude_path, dirpath_for_matching)
            for exclude_path in exclude_paths
        ):
            # Treat excluded paths as invisible
            continue

        if depth:
            # Ignore anything past requested depth
            current_depth = dirpath.count(os.path.sep)
            if current_depth >= base_depth + depth:
                continue
        for filename in filenames:
            if exclude_hidden_paths and filename.startswith("."):
                continue
            file_path = os.path.join(dirpath, filename)
            if filename.endswith(".py"):
                yield file_path


def walk_pypackages(
    root: str,
    depth: Optional[int] = None,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
) -> Generator[str, None, None]:
    for filepath in walk_pyfiles(
        root,
        depth=depth,
        exclude_paths=exclude_paths,
        exclude_hidden_paths=exclude_hidden_paths,
    ):
        init_file_ending = f"{os.path.sep}__init__.py"
        if filepath.endswith(init_file_ending):
            yield filepath[: -len(init_file_ending)]


def walk_configured_packages(
    root: str,
    depth: Optional[int] = None,
    exclude_paths: Optional[list[str]] = None,
    exclude_hidden_paths: Optional[bool] = True,
) -> Generator[str, None, None]:
    for dirpath in walk_pypackages(
        root,
        depth=depth,
        exclude_paths=exclude_paths,
        exclude_hidden_paths=exclude_hidden_paths,
    ):
        package_yml_path = os.path.join(dirpath, "package.yml")
        if os.path.isfile(package_yml_path):
            yield dirpath


@lru_cache(maxsize=None)
def file_to_module_path(file_path: str) -> str:
    # Assuming that the file_path has been 'canonicalized' and does not traverse multiple directories
    file_path = file_path.lstrip("./")
    if file_path == ".":
        return ""

    module_path = file_path.replace(os.sep, ".")

    if module_path.endswith(".py"):
        module_path = module_path[:-3]
    if module_path.endswith(".__init__"):
        module_path = module_path[:-9]
    if module_path == "__init__":
        return ""

    return module_path


def path_exists_case_sensitive(p: Path) -> bool:
    if not p.exists():
        return False

    while True:
        if p == p.parent:
            return True
        # If string representation of path is not in parent directory, return False
        if str(p) not in map(str, p.parent.iterdir()):
            return False
        p = p.parent


def module_to_file_path(
    mod_path: str, find_package_init: bool = False
) -> tuple[str, str]:
    # Assumes that the mod_path is correctly formatted and refers to an actual module
    fs_path = mod_path.replace(".", os.path.sep)

    # mod_path may refer to a package
    if path_exists_case_sensitive(Path(fs_path)):
        return (
            os.path.join(fs_path, "__init__.py") if find_package_init else fs_path
        ), ""

    # mod_path may refer to a file module
    file_path = fs_path + ".py"
    if path_exists_case_sensitive(Path(file_path)):
        return file_path, ""

    # mod_path may refer to a member within a file module
    last_sep_index = fs_path.rfind(os.path.sep)
    file_path = fs_path[:last_sep_index] + ".py"
    if path_exists_case_sensitive(Path(file_path)):
        member_name = fs_path[last_sep_index + 1 :]
        return file_path, member_name

    init_file_path = fs_path[:last_sep_index] + "/__init__.py"
    if path_exists_case_sensitive(Path(init_file_path)):
        member_name = fs_path[last_sep_index + 1 :]
        return init_file_path, member_name

    raise errors.TachParseError(
        f"Failed to translate module path {mod_path} into file path"
    )


if hasattr(sys, "stdlib_module_names"):
    stdlib_module_names = getattr(sys, "stdlib_module_names")
else:
    import stdlib_list

    stdlib_module_names = stdlib_list.stdlib_list()


def is_standard_lib_or_builtin_import(module_base: str) -> bool:
    return module_base in stdlib_module_names or module_base in sys.builtin_module_names


def is_project_import(project_root: str, mod_path: str) -> bool:
    root_base = os.path.basename(os.path.realpath(project_root))
    module_base = mod_path.split(".", 1)[0]
    if is_standard_lib_or_builtin_import(module_base):
        return False
    if root_base == module_base:
        return True
    if os.path.isdir(os.path.join(project_root, module_base)) or os.path.isfile(
        os.path.join(project_root, f"{module_base}.py")
    ):
        return True
    return False



import os

import pytest
from pydantic import ValidationError

from tach.check import check, ErrorInfo
from tach.core.config import PackageConfig, ScopeDependencyRules, ProjectConfig
from tach.parsing.config import parse_project_config, parse_package_config
from tach.filesystem import file_to_module_path
from tach import filesystem as fs


def test_file_to_mod_path():
    assert file_to_module_path("__init__.py") == ""
    assert file_to_module_path("domain_one/__init__.py") == "domain_one"
    assert file_to_module_path("domain_one/interface.py") == "domain_one.interface"


def test_parse_valid_project_config():
    result = parse_project_config("example/valid/")
    assert result == ProjectConfig(
        constraints={
            "one": ScopeDependencyRules(depends_on=["two"]),
            "two": ScopeDependencyRules(depends_on=["one"]),
            "three": ScopeDependencyRules(depends_on=[]),
        },
        exclude=["domain_thr.*"],
        exclude_hidden_paths=True,
    )


def test_run_valid_project_config():
    current_dir = os.getcwd()
    try:
        project = "./example/valid/"
        fs.chdir(project)
        project_config = parse_project_config()
        results = check(
            ".",
            project_config,
            exclude_paths=project_config.exclude,
            exclude_hidden_paths=project_config.exclude_hidden_paths,
        )
        assert results == []
    finally:
        # Make sure not to dirty the test directory state
        fs.chdir(current_dir)


def test_parse_valid_strict_package_config():
    result = parse_package_config("example/valid/domain_one")
    assert result == PackageConfig(strict=True, tags=["one"])


def test_parse_valid_multi_tag_package_config():
    result = parse_package_config("example/valid/domain_two")
    assert result == PackageConfig(strict=False, tags=["two", "shared"])


def test_package_with_no_config():
    result = parse_package_config("example/")
    assert result is None


def test_invalid_project_config():
    with pytest.raises(ValidationError):
        parse_project_config("example/invalid/")


def test_empty_project_config():
    with pytest.raises(ValueError):
        parse_project_config("example/invalid/empty")


def test_invalid_package_config():
    with pytest.raises(ValidationError):
        parse_package_config("example/invalid")


def test_empty_package_config():
    with pytest.raises(ValueError):
        parse_package_config("example/invalid")


def test_exclude_hidden_paths_fails():
    current_dir = os.getcwd()
    hidden_project = "./example/invalid/hidden/"
    fs.chdir(hidden_project)
    try:
        project_config = parse_project_config()
        assert project_config.exclude_hidden_paths is False
        results = check(
            ".",
            project_config,
            exclude_hidden_paths=project_config.exclude_hidden_paths,
        )
        assert len(results) == 1
        assert results[0] == ErrorInfo(
            location="hidden",
            import_mod_path="",
            source_tag="",
            allowed_tags=[],
            exception_message="Package 'unhidden' is in strict mode. Only imports from the root of"
            " this package are allowed. The import 'unhidden.secret.shhhh' (in 'hidden') is not included in __all__.",
        )

        project_config.exclude_hidden_paths = True
        assert check(".", project_config, exclude_hidden_paths=True) == []
    finally:
        # Make sure not to dirty the test directory state
        fs.chdir(current_dir)



import pytest
from tach.core import (
    PackageConfig,
    PackageTrie,
    PackageNode,
    ProjectConfig,
    ScopeDependencyRules,
)
from tach.check import check_import


@pytest.fixture
def test_config() -> PackageConfig:
    return PackageConfig(tags=["test"], strict=False)


@pytest.fixture
def project_config() -> ProjectConfig:
    return ProjectConfig(
        constraints={
            "domain_one": ScopeDependencyRules(
                depends_on=["domain_one", "domain_three"]
            ),
            "domain_two": ScopeDependencyRules(depends_on=["domain_one"]),
            "domain_three": ScopeDependencyRules(depends_on=[]),
        }
    )


@pytest.fixture
def package_trie() -> PackageTrie:
    return PackageTrie(
        root=PackageNode(
            is_end_of_path=False,
            full_path="",
            config=None,
            children={
                "domain_one": PackageNode(
                    is_end_of_path=True,
                    full_path="domain_one",
                    config=PackageConfig(tags=["domain_one"], strict=True),
                    interface_members=["public_fn"],
                    children={
                        "subdomain": PackageNode(
                            is_end_of_path=True,
                            full_path="domain_one.subdomain",
                            config=PackageConfig(tags=["domain_one"], strict=True),
                            children={},
                        )
                    },
                ),
                "domain_two": PackageNode(
                    is_end_of_path=True,
                    full_path="domain_two",
                    config=PackageConfig(tags=["domain_two"], strict=False),
                    children={
                        "subdomain": PackageNode(
                            is_end_of_path=True,
                            full_path="domain_two.subdomain",
                            config=PackageConfig(tags=["domain_two"], strict=False),
                            children={},
                        )
                    },
                ),
                "domain_three": PackageNode(
                    is_end_of_path=True,
                    full_path="domain_three",
                    config=PackageConfig(tags=["domain_three"], strict=False),
                    children={},
                ),
            },
        )
    )


@pytest.mark.parametrize(
    "file_mod_path,import_mod_path,expected_result",
    [
        ("domain_one", "domain_one", True),
        ("domain_one", "domain_one.subdomain", True),
        ("domain_one", "domain_one.core", True),
        ("domain_one", "domain_three", True),
        ("domain_two", "domain_one", True),
        ("domain_two", "domain_one.public_fn", True),
        ("domain_two.subdomain", "domain_one", True),
        ("domain_two", "external", True),
        ("external", "external", True),
        ("domain_two", "domain_one.private_fn", False),
        ("domain_three", "domain_one", False),
        ("domain_two", "domain_one.core", False),
        ("domain_two.subdomain", "domain_one.core", False),
        ("domain_two", "domain_three", False),
        ("domain_two", "domain_two.subdomain", False),
        ("external", "domain_three", False),
    ],
)
def test_check_import(
    project_config, package_trie, file_mod_path, import_mod_path, expected_result
):
    result = check_import(
        project_config=project_config,
        package_trie=package_trie,
        file_mod_path=file_mod_path,
        import_mod_path=import_mod_path,
    )
    assert result.ok == expected_result



from collections import deque
from dataclasses import dataclass, field
from typing import Optional, Generator

from tach.core.config import PackageConfig


@dataclass
class PackageNode:
    """
    A node in the package trie.

    If 'is_end_of_path' is True, this node represents a package in the project,
    and must have 'config' and 'full_path' set.

    If 'is_end_of_path' is False, this node does not represent a real package,
    and must have 'config' None and 'full_path' as the empty string.
    """

    is_end_of_path: bool
    full_path: str
    config: Optional[PackageConfig]
    interface_members: list[str] = field(default_factory=list)
    children: dict[str, "PackageNode"] = field(default_factory=dict)

    @classmethod
    def empty(cls) -> "PackageNode":
        return PackageNode(is_end_of_path=False, full_path="", config=None)

    def fill(
        self, config: PackageConfig, full_path: str, interface_members: list[str]
    ) -> None:
        self.is_end_of_path = True
        self.config = config
        self.full_path = full_path
        self.interface_members = interface_members


@dataclass
class PackageTrie:
    """
    The core data structure for tach, representing the packages in a project
    with a trie structure for package prefix lookups.
    """

    root: PackageNode = field(default_factory=PackageNode.empty)

    def __iter__(self):
        return package_trie_iterator(self)

    @staticmethod
    def _split_mod_path(path: str) -> list[str]:
        # By default "".split(".") -> ['']
        # so we want to remove any whitespace path components
        return [part for part in path.split(".") if part]

    def get(self, path: str) -> Optional[PackageNode]:
        node = self.root
        parts = self._split_mod_path(path)

        for part in parts:
            if part not in node.children:
                return None
            node = node.children[part]

        return node if node.is_end_of_path else None

    def insert(self, config: PackageConfig, path: str, interface_members: list[str]):
        node = self.root
        parts = self._split_mod_path(path)

        for part in parts:
            if part not in node.children:
                node.children[part] = PackageNode.empty()
            node = node.children[part]

        node.fill(config, path, interface_members)

    def find_nearest(self, path: str) -> Optional[PackageNode]:
        node = self.root
        parts = self._split_mod_path(path)
        nearest_parent = node

        for part in parts:
            if part in node.children:
                node = node.children[part]
                if node.is_end_of_path:
                    nearest_parent = node
            else:
                break

        return nearest_parent if nearest_parent.is_end_of_path else None


def package_trie_iterator(trie: PackageTrie) -> Generator[PackageNode, None, None]:
    stack = deque([trie.root])

    while stack:
        node = stack.popleft()
        if node.is_end_of_path:
            yield node

        stack.extend(node.children.values())



from tach.filesystem.service import (
    get_cwd,
    chdir,
    canonical,
    read_file,
    write_file,
    parse_ast,
    walk_pyfiles,
    walk_pypackages,
    walk_configured_packages,
    file_to_module_path,
    module_to_file_path,
    is_project_import,
)
from tach.filesystem.project import (
    get_project_config_path,
    validate_project_config_path,
    print_no_config_yml,
    find_project_config_root,
)
from tach.filesystem.package import (
    validate_package_config,
)

__all__ = [
    "get_cwd",
    "chdir",
    "canonical",
    "read_file",
    "write_file",
    "parse_ast",
    "walk_pyfiles",
    "walk_pypackages",
    "walk_configured_packages",
    "file_to_module_path",
    "module_to_file_path",
    "is_project_import",
    "get_project_config_path",
    "validate_project_config_path",
    "print_no_config_yml",
    "validate_package_config",
    "find_project_config_root",
]



import os
from datetime import datetime
from typing import Iterable, Optional

import yaml

from tach import filesystem as fs
from tach.check import check
from tach.constants import CONFIG_FILE_NAME, PACKAGE_FILE_NAME
from tach.core import ScopeDependencyRules
from tach.errors import TachError
from tach.parsing import parse_project_config, parse_package_config

init_content_template = """# Generated by tach  on {timestamp}
from .main import *
"""


def build_init_content():
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return init_content_template.format(timestamp=timestamp)


def build_package(path: str, tags: Optional[set[str]]) -> Optional[str]:
    dirname = path.removesuffix(".py")
    new_tag = os.path.basename(dirname)
    if not tags:
        tags_to_write = [new_tag]
    else:
        tags_to_write = tags
    if os.path.isfile(path):
        # Create the package directory
        os.mkdir(dirname)
        # Write the __init__
        fs.write_file(os.path.join(dirname, "__init__.py"), build_init_content())
        # Move and rename the file
        os.rename(path, os.path.join(dirname, "main.py"))
    # Write the package.yml
    comma_separated_tags = ",".join(map(lambda tag: f'"{tag}"', tags_to_write))
    package_yml_content = f"tags: [{comma_separated_tags}]\n"
    fs.write_file(
        os.path.join(dirname, f"{PACKAGE_FILE_NAME}.yml"), package_yml_content
    )

    if not tags:
        return new_tag


def update_project_config(root: str, tags: set[str]):
    current_dir = fs.get_cwd()
    try:
        fs.chdir(root)
        project_config = parse_project_config()
        check_errors = check(
            root,
            project_config=project_config,
            exclude_paths=project_config.exclude,
        )
        # TODO: handle case where we pivoted a file in a strict package
        for error in check_errors:
            if error.is_tag_error:
                invalid_tags = set(error.invalid_tags)
                existing_dependencies = set(
                    project_config.constraints.get(
                        error.source_tag, ScopeDependencyRules(depends_on=[])
                    ).depends_on
                )
                if error.source_tag in tags:
                    # This is updating the config for a new tag
                    project_config.constraints[error.source_tag] = ScopeDependencyRules(
                        depends_on=list(existing_dependencies | invalid_tags)
                    )
                if invalid_tags & tags:
                    # This is updating the config for an existing tag
                    project_config.constraints[error.source_tag] = ScopeDependencyRules(
                        depends_on=list(existing_dependencies | (invalid_tags & tags))
                    )

        tach_yml_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
        tach_yml_content = yaml.dump(project_config.model_dump())
        fs.write_file(tach_yml_path, tach_yml_content)

        check_errors = check(
            root, project_config=project_config, exclude_paths=project_config.exclude
        )
        if check_errors:
            return (
                "Could not auto-detect all dependencies, "
                "use 'tach check' to finish initialization manually."
            )
    except Exception as e:
        fs.chdir(current_dir)
        raise e
    fs.chdir(current_dir)


def validate_path(path: str) -> None:
    if not os.path.exists(path):
        raise TachError(f"{path} does not exist.")
    if os.path.isdir(path):
        # 'path' points to a directory
        # so we validate that it is a Python package without an existing package config
        if os.path.exists(
            os.path.join(path, f"{PACKAGE_FILE_NAME}.yml")
        ) or os.path.exists(os.path.join(path, f"{PACKAGE_FILE_NAME}.yaml")):
            raise TachError(f"{path} already contains a {PACKAGE_FILE_NAME}.yml")
        if not os.path.exists(os.path.join(path, "__init__.py")):
            raise TachError(
                f"{path} is not a valid Python package (no __init__.py found)."
            )
    else:
        # 'path' points to a file
        # so we validate that it is a Python file we can 'pivot' to a package
        print(
            f"{path} will be moved into a new package. You may need to update relative imports within this file."
        )
        if not path.endswith(".py"):
            raise TachError(f"{path} is not a Python file.")
        if os.path.exists(path.removesuffix(".py")):
            raise TachError("{path} already has a directory of the same name.")
        dirname = os.path.dirname(path)
        package_config = parse_package_config(dirname)
        if package_config is not None and package_config.strict:
            # this is a file we are pivoting out from a strict package
            # so any of its imports from siblings will start to fail, actually all relative imports will fail
            print(
                f"{path} is contained by a strict package. "
                f"You may need to update imports from '{dirname}' to come through __all__ in __init__.py"
            )
    root = fs.find_project_config_root(path)
    if not root:
        raise TachError(f"{CONFIG_FILE_NAME} does not exist in any parent directories")


def add_packages(paths: set[str], tags: Optional[set[str]]) -> Iterable[str]:
    new_tags: set[str] = set()
    # Validate paths
    for path in paths:
        validate_path(path=path)
    # Build packages
    for path in paths:
        new_tag = build_package(path=path, tags=tags)
        if new_tag:
            new_tags.add(new_tag)
    # Update project config
    project_root = fs.find_project_config_root(path=".")
    if not project_root:
        raise TachError(f"{CONFIG_FILE_NAME} not found.")
    warning = update_project_config(root=project_root, tags=tags if tags else new_tags)
    if warning:
        return [warning]
    return []
